"""Ferda PAIN API."""

from collections.abc import Generator
from json import JSONDecodeError
from typing import Optional

from asgiref.sync import sync_to_async
from django.contrib.auth.models import AbstractUser
from django.http import HttpRequest
from ferda.constants import LOGGER_SESSION_KEY
from httpx import AsyncClient, Auth, HTTPStatusError, Request, RequestError, Response, codes
from ninja import Router, Schema
from pydantic import Field, ValidationError

from ferda_pain.constants import PermissionScopes
from ferda_pain.settings import SETTINGS

router = Router(tags=["PAIN"])


class TokenRequest(Schema):
    """Get JWE token request."""

    permission_scopes: list[PermissionScopes] = Field(
        serialization_alias="scopes", description="PAIN API permission scopes"
    )
    logger_session_id: Optional[str] = Field(
        None, serialization_alias="log", description="FRED logger session identifier"
    )


class TokenResponse(Schema):
    """Get JWE token response."""

    token: str = Field(description="JWE token with permissions")


class ErrorResponse(Schema):
    """Error response details."""

    details: str = Field(description="Error description")


class FixedTokenAuth(Auth):
    """Authentication with fixed token in header."""

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        """Add authentication data to request."""
        if SETTINGS.api_token:
            request.headers["Authorization"] = f"Bearer {SETTINGS.api_token}"
        yield request


def get_permission_scopes(user: AbstractUser) -> set[PermissionScopes]:
    """Get PAIN permission scopes for user."""
    PERMISSION_TO_SCOPES_MAPPING: dict[str, set[PermissionScopes]] = {
        "ferda_pain.can_view": {PermissionScopes.VIEW},
        "ferda_pain.can_view_import": {PermissionScopes.VIEW_IMPORT},
        "ferda_pain.can_edit_account": {PermissionScopes.EDIT_ACCOUNT},
        "ferda_pain.can_edit_payment": {PermissionScopes.EDIT_PAYMENT},
        "ferda_pain.can_edit_prequest": {PermissionScopes.EDIT_PREQUEST},
        "ferda_pain.can_maintain": {PermissionScopes.MAINTAIN},
    }

    result_scopes: set[PermissionScopes] = set()
    for permission, scopes in PERMISSION_TO_SCOPES_MAPPING.items():
        if user.has_perm(permission):
            result_scopes.update(scopes)

    return result_scopes


@router.get(
    "/auth/",
    summary="Get JWE token to authenticate to PAIN admin endpoints",
    response={codes.OK: TokenResponse, codes.BAD_REQUEST: ErrorResponse},
)
async def get_access_token(request: HttpRequest) -> tuple:
    """Get access token with maximum privileges for given user."""
    token_request = TokenRequest(
        permission_scopes=await sync_to_async(get_permission_scopes)(request.user),  # type: ignore[arg-type]
        logger_session_id=await sync_to_async(request.session.get)(LOGGER_SESSION_KEY),
    )
    async with AsyncClient(
        base_url=SETTINGS.api_url,
        verify=SETTINGS.ca_certs,
        auth=FixedTokenAuth(),
        timeout=SETTINGS.api_timeout,
        follow_redirects=True,
    ) as client:
        # PAIN API call
        try:
            response = await client.post("token/", json=token_request.model_dump(by_alias=True))
            response.raise_for_status()
        except (HTTPStatusError, RequestError) as e:
            return codes.BAD_REQUEST, {"details": f"Call to PAIN: {e}"}

        # Validate response
        try:
            token_response = TokenResponse(**response.json())
        except (JSONDecodeError, ValidationError):
            return codes.BAD_REQUEST, {"details": "Call to PAIN: Unexpected response"}

        return codes.OK, token_response
