"""Tests of API endpoints."""

import respx
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser, Permission
from django.test import TestCase, override_settings
from django.urls import reverse
from ferda.tests.utils import LOGGER_SESSION_ID, LoggerMixin
from httpx import ConnectTimeout, Response

from ferda_pain.api import get_permission_scopes
from ferda_pain.constants import PermissionScopes

User = get_user_model()


class GetPermissionScopesTest(LoggerMixin, TestCase):
    """Test of `get_permission_scopes` function."""

    def _test_case(
        self, user: AbstractUser, permissions: list[str], expected_scopes: set[PermissionScopes]
    ) -> None:
        user.user_permissions.set(
            Permission.objects.filter(codename__in=permissions, content_type__model="ferda_pain")
        )
        self.client.force_login(user)

        self.assertEqual(get_permission_scopes(user), expected_scopes)

    def test_permissions(self):
        data: list[tuple[list[str], set[PermissionScopes]]] = [
            ([], set()),
            (["can_view"], {PermissionScopes.VIEW}),
            (["can_view_import"], {PermissionScopes.VIEW_IMPORT}),
            (["can_edit_account"], {PermissionScopes.EDIT_ACCOUNT}),
            (["can_edit_payment"], {PermissionScopes.EDIT_PAYMENT}),
            (["can_edit_prequest"], {PermissionScopes.EDIT_PREQUEST}),
            (["can_maintain"], {PermissionScopes.MAINTAIN}),
            (
                ["can_view", "can_edit_account", "can_maintain", "can_view"],
                {PermissionScopes.VIEW, PermissionScopes.EDIT_ACCOUNT, PermissionScopes.MAINTAIN},
            ),
        ]

        for permissions, scopes in data:
            user = User.objects.create_user("test")
            with self.subTest(permissions=permissions, scopes=scopes):
                self._test_case(user, permissions, scopes)
            user.delete()

    def test_superuser(self):
        user = User.objects.create_superuser("admin")
        self._test_case(user, [], set(PermissionScopes))


@override_settings(
    ROOT_URLCONF="ferda_pain.tests.urls",
    FERDA_PAIN_API_URL="https://pain.example.org/api/",
    FERDA_PAIN_API_TOKEN="FIXED_API_TOKEN",
)
class GetAccessTokenTest(LoggerMixin, TestCase):
    """Tests of `get_access_token` endpoint."""

    @respx.mock
    def test_error(self):
        user = User.objects.create_user("test")
        self.client.force_login(user)
        respx.post("https://pain.example.org/api/token/").mock(
            side_effect=ConnectTimeout("Timed out")
        )

        response = self.client.get(reverse("ferda_pain:api:get_access_token"))

        self.assertEqual(response.status_code, 400)
        self.assertJSONEqual(response.content, {"details": "Call to PAIN: Timed out"})

    @respx.mock
    def test_bad_status(self):
        user = User.objects.create_user("test")
        self.client.force_login(user)
        respx.post("https://pain.example.org/api/token/").mock(
            return_value=Response(401, text="Error")
        )

        response = self.client.get(reverse("ferda_pain:api:get_access_token"))

        self.assertEqual(response.status_code, 400)
        self.assertIn("Call to PAIN: Client error '401 Unauthorized'", response.json()["details"])

    @respx.mock
    def test_not_json_response(self):
        user = User.objects.create_user("test")
        self.client.force_login(user)
        respx.post("https://pain.example.org/api/token/").mock(
            return_value=Response(200, text="Not a valid JSON {")
        )

        response = self.client.get(reverse("ferda_pain:api:get_access_token"))

        self.assertEqual(response.status_code, 400)
        self.assertJSONEqual(response.content, {"details": "Call to PAIN: Unexpected response"})

    @respx.mock
    def test_unexpected_response(self):
        user = User.objects.create_user("test")
        self.client.force_login(user)
        respx.post("https://pain.example.org/api/token/").mock(
            return_value=Response(200, json={"field": "value"})
        )

        response = self.client.get(reverse("ferda_pain:api:get_access_token"))

        self.assertEqual(response.status_code, 400)
        self.assertJSONEqual(response.content, {"details": "Call to PAIN: Unexpected response"})

    @respx.mock
    def test_ok(self):
        user = User.objects.create_user("test")
        user.user_permissions.set(
            Permission.objects.filter(codename="can_view", content_type__model="ferda_pain")
        )
        self.client.force_login(user)
        respx.post("https://pain.example.org/api/token/").mock(
            return_value=Response(200, json={"token": "TEST_JWE"})
        )

        response = self.client.get(reverse("ferda_pain:api:get_access_token"))

        # Check response
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {"token": "TEST_JWE"})
        # Check call
        self.assertJSONEqual(
            respx.calls.last.request.content,
            {"log": LOGGER_SESSION_ID, "scopes": [PermissionScopes.VIEW]},
        )
        self.assertEqual(
            respx.calls.last.request.headers["Authorization"], "Bearer FIXED_API_TOKEN"
        )

    @override_settings(FERDA_PAIN_API_TOKEN=None)
    @respx.mock
    def test_ok_no_auth(self):
        user = User.objects.create_user("test")
        self.client.force_login(user)
        respx.post("https://pain.example.org/api/token/").mock(
            return_value=Response(200, json={"token": "TEST_JWE"})
        )

        response = self.client.get(reverse("ferda_pain:api:get_access_token"))

        self.assertEqual(response.status_code, 200)
        self.assertIsNone(respx.calls.last.request.headers.get("Authorization"))
